#include "Scene.h"
#include "../thirdparty/OBJ_Loader.h"
#include "Sphere.h"
#include <QMatrix4x4>
#include <QQuaternion>

namespace raytry {
    Scene::Scene() {
        background.emplace_back(ground);
        for (int i = 0; i < spheres.size(); ++i) {
            float x = -5.f + static_cast<float>(i);
            float spec = std::pow(2.0f, static_cast<float>(i));
            auto mat = RenderMaterial{{197, 52, 27}, {240, 240, 240}, spec};
            spheres[i] = Sphere(camera.pos() + camera.fwd() * 5 + QVector3D(x, 0.f, 0.f), 0.4f, mat);
            background.emplace_back(spheres[i]);
        }
        const auto &triangleTransformation = camera.pos() + camera.fwd() * 10;

        const auto &triangleBase = camera.pos() + camera.fwd() * 10;
        triangles[0] = Triangle(QVector3D{triangleBase[0] - 3, triangleBase[1], 0.0},
                                QVector3D{triangleBase[0] + 3, triangleBase[1], 0.0},
                                QVector3D{triangleBase[0], triangleBase[1], 6.0});
        background.emplace_back(triangles[0]);

        QMatrix4x4 transformation{};
        transformation.translate(camera.pos() + (camera.fwd() * 5));
        transformation.rotate(QQuaternion::fromAxisAndAngle(1.0f, 0.0f, 0.0f, 90.0f));
        transformation.rotate(QQuaternion::fromAxisAndAngle(0.0f, 1.0f, 0.0f, 90.0f));

        std::setlocale(LC_ALL, "C");// the obj parser uses locale-specific number format...
        objl::Loader objLoader{};
        if (!objLoader.LoadFile("newell_teaset/teapot.obj")) {
            qWarning() << "Could not load teapot..";
            return;
        } else {
            auto mat = RenderMaterial{{37, 96, 185}};
            teapotMesh.reserve(teapotMesh.size() + (objLoader.LoadedIndices.size() / 3));

            auto iter = objLoader.LoadedIndices.begin();
            while (iter != objLoader.LoadedIndices.end()) {
                auto a = objLoader.LoadedVertices[*iter++];
                auto b = objLoader.LoadedVertices[*iter++];
                auto c = objLoader.LoadedVertices[*iter++];
                QVector3D qta{a.Position.X, a.Position.Y, a.Position.Z};
                QVector3D qtan{a.Normal.X, a.Normal.Y, a.Normal.Z};
                QVector3D qtb{b.Position.X, b.Position.Y, b.Position.Z};
                QVector3D qtbn{b.Normal.X, b.Normal.Y, b.Normal.Z};
                QVector3D qtc{c.Position.X, c.Position.Y, c.Position.Z};
                QVector3D qtcn{c.Normal.X, c.Normal.Y, c.Normal.Z};
                Triangle &tri = teapotMesh.emplace_back(
                        transformation.map(qta),
                        transformation.map(qtb),
                        transformation.map(qtc),
                        transformation.mapVector(qtan),
                        transformation.mapVector(qtbn),
                        transformation.mapVector(qtcn),
                        mat);
                //tri.setMaterial(mat);
            }
        }
        qInfo() << "Loaded teapot with" << teapotMesh.size() << "triangles.";

        foregroundTree = KD::Tree::build(teapotMesh);
    }

    std::list<std::pair<std::reference_wrapper<const RenderObject>, float>> Scene::findIntersections(const Ray &ray, IntersectionMode mode) {
        std::list<std::pair<std::reference_wrapper<const RenderObject>, float>> matches{};
        if ((mode == IntersectionMode::foreground || mode == IntersectionMode::full) && !foregroundTree.getNodes().empty()) {
            const QVector3D invDirection{1 / ray.directionVec.x(), 1 / ray.directionVec.y(), 1 / ray.directionVec.z()};
            const std::optional<std::pair<QVector3D, QVector3D>> &treeHit = foregroundTree.getBounds().fastslabtest(ray, invDirection);
            if (treeHit) {
                // KD!
                const KD::Node *node = foregroundTree.getNodes().data();
                float tMin = treeHit->first[node->getAxis()];
                float tMax = treeHit->second[node->getAxis()];
                unsigned int currentNodeIdx = 0;
                std::list<std::tuple<unsigned int, float, float>> nodes{};
                while (node != nullptr) {
                    if (node->isLeaf()) {
                        assert(node->farIdx() >= 0);
                        assert(node->farIdx() < foregroundTree.getLeafs().size());
                        for (const auto &item: foregroundTree.getLeafs()[node->farIdx()].triangles) {
                            const auto distance = item->intersects(ray);
                            ++triangleIntersections;
                            if (distance) {
                                bool emplaced = false;
                                const RenderObject &triangle = *item;
                                for (auto mpos = matches.begin(); !emplaced && mpos != matches.end(); ++mpos) {
                                    if ((*mpos).second > distance.value()) {
                                        matches.emplace(mpos, triangle, distance.value());
                                        emplaced = true;
                                    }
                                }
                                if (!emplaced) {
                                    matches.emplace_back(triangle, distance.value());
                                }
                            }
                        }

                        if (nodes.empty()) {
                            node = nullptr;
                        } else {
                            std::tie(currentNodeIdx, tMin, tMax) = nodes.front();
                            assert(currentNodeIdx < foregroundTree.getNodes().size());
                            node = &foregroundTree.getNodes()[currentNodeIdx];
                            nodes.pop_front();
                        }
                    } else {
                        const auto tPlane = (node->val() - ray.originVec[node->getAxis()]) * invDirection[node->getAxis()];

                        unsigned int firstChild, secondChild;
                        int belowFirst = (ray.originVec[node->getAxis()] < node->val()) ||
                                         (ray.originVec[node->getAxis()] == node->val() && ray.directionVec[node->getAxis()] <= 0);
                        if (belowFirst) {
                            firstChild = currentNodeIdx + 1;
                            secondChild = currentNodeIdx + node->farIdx();
                        } else {
                            firstChild = currentNodeIdx + node->farIdx();
                            secondChild = currentNodeIdx + 1;
                        }

                        if (tPlane > tMax || tPlane <= 0) {
                            currentNodeIdx = firstChild;
                        } else if (tPlane < tMin) {
                            currentNodeIdx = secondChild;
                        } else {
                            nodes.emplace_back(secondChild, tPlane, tMax);
                            currentNodeIdx = firstChild;
                            tMax = tPlane;
                        }
                        node = &foregroundTree.getNodes()[currentNodeIdx];
                    }
                }
            }
        }

        if (mode == IntersectionMode::background || mode == IntersectionMode::full) {
            for (RenderObject &obj: background) {
                const auto distance = obj.intersects(ray);
                if (distance) {
                    bool emplaced = false;
                    for (auto mpos = matches.begin(); !emplaced && mpos != matches.end(); ++mpos) {
                        if ((*mpos).second > distance.value()) {
                            matches.emplace(mpos, obj, distance.value());
                            emplaced = true;
                        }
                    }
                    if (!emplaced) {
                        matches.emplace_back(obj, distance.value());
                    }
                }
            }
        }
        return matches;
    }

    QQuaternion Scene::nextRandomRotation() {
        auto r1 = pipiDistribution(randomGenerator);
        auto r2 = halfpiDistribution(randomGenerator);
        return QQuaternion::rotationTo({0.0f, 0.0f, 1.0f}, {std::sin(r1) * std::sin(std::acos(r2)), std::cos(r1) * std::sin(std::acos(r2)), r2});
    }

    QColor Scene::incomingLight(const QVector3D &point, const QVector3D &incomingDir, unsigned int depth) {
        const Ray &lightPathRay{point, -incomingDir};
        // only background for performance reasons
        auto matches = findIntersections(lightPathRay, IntersectionMode::background);
        if (matches.empty()) {
            float lightIntensity = ((point - pointLight).normalized() - incomingDir.normalized()).lengthSquared();
            if (lightIntensity <= 1) {
                return mixColor(lightColor, lightIntensity);
            } else {
                return QColorConstants::Black;
            }
        } else {
            auto [object, distance] = matches.front();
            QVector3D hitpoint = lightPathRay.originVec + (lightPathRay.directionVec * distance);
            return renderingEquation(object, hitpoint, -lightPathRay.directionVec, depth + 1);
        }
    }

    QColor Scene::renderingEquation(const RenderObject &object, const QVector3D &hitpoint, const QVector3D &outgoingDir, unsigned int depth) {
        if (depth > maxDepth) {
            return QColorConstants::Black;
        }

        const RenderMaterial &material = object.material();
        const QVector3D &normal = object.calculateNormal(hitpoint);
        const QVector3D &toLight = (pointLight - hitpoint).normalized();

        // multiple directions on first intersection/depth
        if (depth == 0u) {
            QColor sum = QColor{0, 0, 0};
            float dW = 1.0f / static_cast<float>(pathRandomDirectionCount);
            for (unsigned int i = 0; i < pathRandomDirectionCount; ++i) {
                auto Wi = -nextRandomRotation().rotatedVector(normal);
                sum = addColors(sum, mixColor(addColors(material.brdf(toLight, normal, Wi, outgoingDir), incomingLight(hitpoint, Wi, depth)), std::max(0.0f, -QVector3D::dotProduct(normal, Wi)) * dW));
            }
            return sum;
        } else {
            auto Wi = -nextRandomRotation().rotatedVector(normal);
            return mixColor(addColors(material.brdf(toLight, normal, Wi, outgoingDir), incomingLight(hitpoint, Wi, depth)), std::max(0.0f, -QVector3D::dotProduct(normal, Wi)));
        }
    }

    void Scene::render(ViewerWindow &window, std::function<void()> updateImage) {
        size_t bundleC = 0;
        QImage &img = window.getDisplayImage();
        unsigned int pixels = img.width() * img.height();

#ifdef TESTING
        QColor color = addColors(QColorConstants::White, QColorConstants::White);
        qInfo() << "Added colors:" << color;
        assert(color == QColorConstants::White);
        color = mixColors(QColorConstants::White, QColorConstants::White);
        qInfo() << "Mixed colors:" << color;
        assert(color == QColorConstants::White);
#endif

        camera.for_each_bundle(img, [&](int x, int y, const std::array<Ray, bundle_size> &bundle) {
            ++bundleC;
            if (bundleC % std::max(1u, pixels / 20) == 0) {
                qInfo() << "Pathtracing #" << bundleC << "/" << pixels << "(" << bundleC * 100 / pixels << "%)";
            }

            QColor mixedColor{QColorConstants::Black};
            for (const auto &ray: bundle) {
                QColor rayColor;
                auto matches = findIntersections(ray, full);
                if (matches.empty()) {
                    rayColor = backgroundColor;
                } else {
                    auto [object, distance] = matches.front();
                    QVector3D hitpoint = ray.originVec + (ray.directionVec * distance);
                    rayColor = renderingEquation(object, hitpoint, -ray.directionVec, 0);
                }
                mixedColor = addColors(mixedColor, mixColor(rayColor, 1.0f / bundle_size));
            }
            img.setPixelColor(x, y, mixedColor);

            if (bundleC % std::max(1u, pixels / 1000) == 0) {
                updateImage();
            }
        });
        qInfo() << "All pixels computed. Tested" << triangleIntersections << "ray-triangle intersections.";
    }
}// namespace raytry
